'use strict';

module.exports = function (_ref) {
  var t = _ref.types;

  var decodeDefineArguments = function decodeDefineArguments(argNodes) {
    if (argNodes.length === 1) {
      return { factory: argNodes[0] };
    } else if (argNodes.length === 2) {
      var decodedArgs = { factory: argNodes[1] };
      if (t.isArrayExpression(argNodes[0])) {
        decodedArgs.dependencyList = argNodes[0];
      }
      return decodedArgs;
    } else {
      return { dependencyList: argNodes[1], factory: argNodes[2] };
    }
  };

  var decodeRequireArguments = function decodeRequireArguments(argNodes) {
    if (argNodes.length === 1) {
      return { dependencyList: argNodes[0] };
    } else {
      return { dependencyList: argNodes[0], factory: argNodes[1] };
    }
  };

  var createModuleExportsAssignmentExpression = function createModuleExportsAssignmentExpression(value) {
    return t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.identifier('module'), t.identifier('exports')), value));
  };

  var createRequireExpression = function createRequireExpression(dependencyNode, variableName) {
    var requireCall = t.callExpression(t.identifier('require'), [dependencyNode]);
    if (variableName) {
      return t.variableDeclaration('var', [t.variableDeclarator(variableName, requireCall)]);
    } else {
      return t.expressionStatement(requireCall);
    }
  };

  return {
    visitor: {
      ExpressionStatement: function ExpressionStatement(path) {
        var node = path.node,
            parent = path.parent;


        if (!t.isCallExpression(node.expression)) return;

        var name = node.expression.callee.name;


        if (name === 'define' && !t.isProgram(parent)) return;

        var argumentDecoders = {
          define: decodeDefineArguments,
          require: decodeRequireArguments
        };
        var argumentDecoder = argumentDecoders[name];

        if (!argumentDecoder) return;

        var _argumentDecoder = argumentDecoder(node.expression.arguments),
            dependencyList = _argumentDecoder.dependencyList,
            factory = _argumentDecoder.factory;

        if (!t.isArrayExpression(dependencyList) && !t.isFunctionExpression(factory)) return;

        var injectsModuleOrExports = false;
        var requireExpressions = [];

        if (dependencyList) {
          dependencyList.elements.forEach(function (el, i) {
            if (t.isStringLiteral(el)) {
              switch (el.value) {
                case 'require':
                  return;
                case 'module':
                case 'exports':
                  injectsModuleOrExports = true;
                  return;
              }
            }
            var paramName = factory && factory.params[i];
            requireExpressions.push(createRequireExpression(el, paramName));
          });
        }

        if (factory) {
          var factoryArity = factory.params.length;
          var replacementFuncExpr = t.functionExpression(null, [], t.blockStatement(requireExpressions.concat(factory.body.body)));
          var replacementCallExprParams = [];

          // https://github.com/requirejs/requirejs/wiki/differences-between-the-simplified-commonjs-wrapper-and-standard-amd-define
          var isSimplifiedCommonJSWrapper = !dependencyList && factoryArity > 0;
          if (isSimplifiedCommonJSWrapper) {
            replacementFuncExpr = factory;
            var identifiers = ['require', 'exports', 'module'];
            replacementCallExprParams = identifiers.slice(0, factoryArity).map(function (a) {
              return t.identifier(a);
            });
          }

          var factoryReplacement = t.callExpression(replacementFuncExpr, replacementCallExprParams);

          injectsModuleOrExports = injectsModuleOrExports || !dependencyList && factoryArity > 1;
          if (name === 'define' && !injectsModuleOrExports) {
            path.replaceWith(createModuleExportsAssignmentExpression(factoryReplacement));
          } else {
            path.replaceWith(factoryReplacement);
          }
        } else {
          path.replaceWithMultiple(requireExpressions);
        }
      }
    }
  };
};
